
import requests
import sys
import os
import yaml
import json
import asyncio

from toledo.utils import set_user_agent
from toledo.utils import clear_session

from toledo.utils import get_start_html
from toledo.utils import get_saml2_relay_and_request
from toledo.utils import get_saml2_relay_and_response
from toledo.utils import get_saml2_csrf
from toledo.utils import get_data_account_id
from toledo.utils import get_websocket_payload
from toledo.utils import get_vivesplus_auth_token
from toledo.utils import next_auth_provoke

from toledo.utils import post_saml2_relay_and_request
from toledo.utils import post_saml2_csrf
from toledo.utils import post_saml2_credentials
from toledo.utils import check_login_state

from toledo.utils import SESSION


class VivesPlusLogin:

    # This class retrieves the vives plus auth token

    def __init__(self, user: str, password: str) -> None:

        try:

            with open(os.path.join(os.path.dirname(__file__), 'config.yaml'), 'r') as f:
                parser = yaml.safe_load(f)

        except FileNotFoundError:

            sys.exit('Unable to find find config.yaml')

        self._PORTALURL = parser['LOGIN']['PortalURLMobile']

        self._AUTHORIZATION_ENDPOINT = parser['LOGIN']['AuthorizationEndpoint']
        self._TOKEN_ENDPOINT = parser['LOGIN']['TokenEndpoint']
        self._BROKER_ENDPOINT = parser['LOGIN']['BrokerEndpoint']

        self._USER = user
        self._PASSWORD = password
        self._USER_AGENT = parser['USER']['UserAgentMobile']

    def __call__(self) -> requests.Session:

        try:

            if '' in (self._USER, self._PASSWORD):

                raise Exception('Please check your credentials!')

            # 1. Vives Plus

            # Clear session
            clear_session()

            # Set User-Agent
            set_user_agent(
                user_agent=self._USER_AGENT
            )

            # Load Toledo Portal
            portal_html = get_start_html(
                url=self._PORTALURL
            )

            # Get RelayState & SAMLRequest
            relaystate_samlrequest = get_saml2_relay_and_request(
                html=portal_html
            )

            # Post RelayState & SAMLRequest
            first_csrf_html = post_saml2_relay_and_request(
                post_info=relaystate_samlrequest
            )

            # Get first csrf token
            first_csrf = get_saml2_csrf(
                html=first_csrf_html
            )

            # Post first csrf token
            second_csrf_html = post_saml2_csrf(
                post_info=first_csrf
            )

            # Get second csrf token
            second_csrf = get_saml2_csrf(
                html=second_csrf_html
            )

            # Post credentials and second csrf token
            third_csrf_html = post_saml2_credentials(
                post_info=second_csrf,
                username=self._USER,
                password=self._PASSWORD
            )

            # Check login state
            if not check_login_state(html=third_csrf_html):
                raise Exception('Invalid credentials')

            # Get data-account-id (nextauthAccountId)
            data_account_id_info = get_data_account_id(
                html=third_csrf_html
            )

            # Get third csrf token
            third_csrf = get_saml2_csrf(
                html=third_csrf_html
            )

            # Get WebSocket Payload
            websocket_payload_data = get_websocket_payload(
                html=third_csrf_html
            )

            # Get fourth csrf token
            fourth_csrf = asyncio.run(next_auth_provoke(
                payload_data=websocket_payload_data,
                data_account_id_info=data_account_id_info,
                post_info=third_csrf
            ))

            # Post fourth csrf token
            fifth_csrf_html = post_saml2_csrf(
                post_info=fourth_csrf
            )

            # Get fifth csrf token
            fifth_csrf = get_saml2_csrf(
                html=fifth_csrf_html
            )

            # Post fifth csrf token
            relaystate_samlresponse_html = post_saml2_csrf(
                post_info=fifth_csrf,
                event_id='proceed'
            )

            # Get RelayState & SAMLResponse
            relaystate_samlresponse = get_saml2_relay_and_response(
                html=relaystate_samlresponse_html
            )

            if relaystate_samlresponse is None:

                raise Exception(
                    'Please create a recovery code before using this package!')

            # Post RelayState & SAMLResponse
            get_vivesplus_auth_token(
                post_info=relaystate_samlresponse
            )

            return SESSION

        except requests.exceptions.HTTPError as errh:
            sys.exit("Http Error:", errh)
        except requests.exceptions.ConnectionError as errc:
            sys.exit("Error Connecting:", errc)
        except requests.exceptions.Timeout as errt:
            sys.exit("Timeout Error:", errt)
        except requests.exceptions.RequestException as err:
            sys.exit("OOps: Something Else: ", err)
        except Exception as ex:
            sys.exit(f"Error: {ex}")


def create_vivesplus_session_object(user: str = None, password: str = None, authorization: str = None) -> requests.Session:

    try:

        if authorization is None and None not in (user, password):

            return VivesPlusLogin(
                user=user,
                password=password
            )()

        elif authorization is not None and None in (user, password):
            
            if os.path.isfile(authorization):

                with open(authorization, 'r') as f:

                    token = json.load(f)['token']

            else:

                if 'Bearer' not in authorization:

                    raise Exception('An authorization token should contain "Bearer ..."!') 

                token = authorization

            session = requests.Session()
            session.headers.update(
                {
                    'Authorization': token
                }
            )
            return session

        else:

            raise Exception('Please provide (username/password) or (authorization token/file)')
    
    except Exception as ex:

        sys.exit(f'Credentials Error: {ex}')



import json
import requests
import re
import html as html_lib
import websockets

from bs4 import BeautifulSoup
from urllib.parse import parse_qs, urlparse

''' SESSION CONFIG '''

SESSION = requests.Session()

def set_user_agent(user_agent: str) -> None:

    SESSION.headers.update({
        'User-Agent': user_agent
    })

def clear_session() -> None:

    SESSION.headers.clear()
    SESSION.cookies.clear()


''' SAML2 GET METHODS '''


def get_start_html(url: str) -> str:
    ''' HTTP GET to portal, dashboard or kuloket url
        Returns inital HTML
    '''

    r = SESSION.get(
        url=url
    )

    r.raise_for_status()

    return r.text


def get_saml2_relay_and_request(html: str) -> dict:
    ''' Searches and returns url, relaysate and samlrequest values from provided html'''

    soup = BeautifulSoup(html, 'html.parser')

    form = soup.body.form

    return {

        'url': form['action'],
        'relaystate': form.find('input', {'name': 'RelayState'})['value'] if form.find('input', {'name': 'RelayState'}) is not None else None ,
        'samlrequest': form.find('input', {'name': 'SAMLRequest'})['value'],
    }


def get_saml2_relay_and_response(html: str) -> dict:
    ''' Searches and returns url, relaysate and samlresponse values from provided html
        Returns None when the relaystate can't be found
    '''

    soup = BeautifulSoup(html, 'html.parser')

    form = soup.body.form

    if form.find('input', {'name': 'RelayState'}) is not None:

        return {

            'url': form['action'],
            'relaystate': form.find('input', {'name': 'RelayState'})['value'],
            'samlresponse': form.find('input', {'name': 'SAMLResponse'})['value'],
        }
    
    elif form.find('input', {'name': 'SAMLResponse'}) is not None:

        return {

            'url': form['action'],
            'samlresponse': form.find('input', {'name': 'SAMLResponse'})['value'],
        }

    else:

        return None


def get_saml2_csrf(html: str) -> dict:
    ''' Searches and returns url and csrf token from provided html'''

    soup = BeautifulSoup(html, 'html.parser')

    form = soup.find('form')

    url = f"https://idp.kuleuven.be{form['action']}" if 'SAML2' in form['action'] else f"https://webwsp.aps.kuleuven.be{form['action']}"

    return {
        'url': url,
        'csrf_token': form.find('input', {'name': 'csrf_token'})['value']
    }


def get_saml2_session_ss(html: str) -> str:
    ''' Find session_ss in html and return it'''
    soup = BeautifulSoup(html, 'html.parser')

    scripts = soup.find_all('script')

    session_ss_raw = [
        s.string for s in scripts if 'writeLocalStorage("shib_idp_session_ss"' in s.string][0]

    session_ss = re.search(r'(shib_idp_session_ss",)(.*)"\);',
                           session_ss_raw).group(2)

    return session_ss


''' SAML2 POST METHODS'''


def post_saml2_relay_and_request(post_info: dict) -> str:
    ''' HTTP POST relaystate and samlrequest and return new html'''

    r = SESSION.post(
        url=post_info['url'],
        data={
            'RelayState': post_info['relaystate'],
            'SAMLRequest': post_info['samlrequest']
        }
    )

    r.raise_for_status()

    return r.text


def post_saml2_relay_and_response(post_info: dict, return_code: bool = False) -> dict:
    ''' HTTP POST relaystate and samlresponse and return code'''

    url = f"https://webwsp.aps.kuleuven.be{post_info['url']}" if '/sap/bc/ui2/flp' in post_info['url'] else post_info['url']

    r = SESSION.post(
        url=url,
        data={
            'RelayState': post_info['relaystate'],
            'SAMLResponse': post_info['samlresponse']
        }
    )

    r.raise_for_status()

    if return_code:

        query = parse_qs(urlparse(r.url).query)

        return query['code'][0]

    else:

        return r.text


def post_saml2_csrf(post_info: dict, session_ss: str = '', event_id: str = '') -> str:
    ''' HTTP POST csrf token and return html'''
    
    url = f'https://idp.kuleuven.be{post_info["url"]}' if 'https://idp.kuleuven.be' not in post_info["url"] else post_info["url"] 

    r = SESSION.post(
        url=url,
        data={
            'csrf_token': post_info['csrf_token'],
            'shib_idp_ls_exception.shib_idp_session_ss': '',
            'shib_idp_ls_success.shib_idp_session_ss': 'true',
            'shib_idp_ls_value.shib_idp_session_ss': session_ss,
            'shib_idp_ls_exception.shib_idp_persistent_ss': '',
            'shib_idp_ls_success.shib_idp_persistent_ss': 'true',
            'shib_idp_ls_value.shib_idp_persistent_ss': '',
            'shib_idp_ls_supported': 'true',
            '_eventId_proceed': event_id

        }
    )
    
    r.raise_for_status()

    if r.history:
        
        shib_idp_session_cookie = r.history[0].cookies.get_dict()['shib_idp_session'] if 'shib_idp_session' in r.history[0].cookies.get_dict() else None
        
        SESSION.cookies.update({'shib_idp_session': shib_idp_session_cookie})
    
    return r.text


def post_saml2_credentials(post_info: dict, username: str, password: str) -> str:
    ''' HTTP POST csrf token and credentials and return html'''

    r = SESSION.post(
        url=post_info['url'],
        data={
            'csrf_token': post_info['csrf_token'],
            'username': username,
            'password': password,
            '_eventId': 'proceed'

        }
    )

    r.raise_for_status()

    return r.text

def post_saml2_nextauth(data_account_id_info: dict, post_info: dict) -> str:
    ''' HTTP POST csrf token and data-account-id and return html'''

    r = SESSION.post(
        url=post_info['url'],
        data={
            'csrf_token': post_info['csrf_token'],
            '_eventId_ProvokeLoginOnAccount': '',
            'nextauthAccountId': data_account_id_info['data-account-id']
        },
        headers={
            'X-Requested-With': 'XMLHttpRequest'
        }
    )

    r.raise_for_status()

    return r.json()


''' WEBSOCKET METHODS '''

async def next_auth_provoke(payload_data: dict, data_account_id_info: dict, post_info: dict):
    
    async with websockets.connect(payload_data['url']) as websocket:

        await websocket.send(f'REGISTER {payload_data["payload"]}')
        
        fourth_csrf = post_saml2_nextauth(data_account_id_info=data_account_id_info, post_info=post_info)
        
        while True:
            
            if 'LOGIN' in await websocket.recv():
                break
            
        return fourth_csrf
    

''' VIVES PLUS METHODS '''

def get_vivesplus_auth_token(post_info: dict) -> str:
    ''' HTTP POST samlresponse and return Vives Plus auth'''

    r = SESSION.post(
        url=post_info['url'],
        data={
            'SAMLResponse': post_info['samlresponse']
        },
    )

    r.raise_for_status()

    SESSION.headers.update({
        'Authorization': f"Bearer {r.json()['id_token']}"
    })


''' OTHER METHODS '''


def openid_connect_auth(state: str, code_challenge: str, url: str) -> str:
    ''' HTTP GET with sate and code_challenge as parameters
        Returns HTML with csrf token
    '''

    r = SESSION.get(
        url=url,
        params={
            'client_id': 'toledo-la-dashboard',
            'redirect_uri': 'https://toledo.kuleuven.be/dashboard/auth/signinwin/kul',
            'response_type': 'code',
            'scope': 'openid profile email',
            'state': state,
            'code_challenge': code_challenge,
            'code_challenge_method': 'S256',
            'response_mode': 'query'
        },
        allow_redirects=True
    )

    return r.text


def get_dashboard_auth_token(code: str, code_verifier: str, url: str) -> dict:
    ''' HTTP POST with code and code_verifier
        Returns dashboard access token as json
    '''

    r = SESSION.post(
        url=url,
        data={
            'client_id': 'toledo-la-dashboard',
            'code': code,
            'redirect_uri': 'https://toledo.kuleuven.be/dashboard/auth/signinwin/kul',
            'code_verifier': code_verifier,
            'grant_type': 'authorization_code'
        }
    )

    r.raise_for_status()

    SESSION.headers.update({
        'Authorization': f"Bearer {json.loads(r.text)['access_token']}"
    })


def get_start_up_info(html: str) -> dict:

    soup = BeautifulSoup(html, 'html.parser')

    meta = soup.find('meta', {'name': 'sap.ushellConfig.serverSideConfig.1'})
    content_str = html_lib.unescape(meta['content'])
    content_json = json.loads(content_str)

    return content_json['startupConfig']['services']['startUp']


def get_schedule_manifest(url_info: dict, schedule_url: str) -> str:

    so = urlparse(schedule_url).path.split('/kuloket/')[1]

    r = SESSION.get(
        url=f"https://webwsp.aps.kuleuven.be{url_info['baseUrl']}{url_info['relativeUrl']}",
        params={
            'so': so,
            'action': 'display',
            'formFactor': 'desktop',
            'shellType': 'FLP',
            'depth': 0
        }
    )

    r.raise_for_status()

    targetmappings = json.loads(r.text)['targetMappings']
    key = list(targetmappings.keys())[0]
    info = targetmappings[key]['applicationDependencies']
    manifest = eval(info)
    return f"https://webwsp.aps.kuleuven.be{manifest['manifest']}"


def get_schedule_call_url(url: str) -> str:

    r = SESSION.get(
        url=url,
        params={
            'sap-language': 'EN',
            'sap-client': 200

        }
    )

    r.raise_for_status()

    return f"https://webwsp.aps.kuleuven.be{json.loads(r.text)['sap.app']['dataSources']['mainService']['uri']}"


def get_x_csrf_token(url: str) -> str:

    r = SESSION.head(
        url=url,
        headers={
            'x-csrf-token': 'Fetch',
            'X-Requested-With': 'XMLHttpRequest',
            'X-XHR-Logon': 'accept="iframe,strict-window,window"'

        }
    )

    r.raise_for_status()

    SESSION.headers.update({
        'x-csrf-token': r.headers['x-csrf-token'],
        'X-Requested-With': 'XMLHttpRequest',
        'X-XHR-Logon': 'accept="iframe,strict-window,window"'
    })

def get_data_account_id(html: str) -> dict:
    ''' Searches and returns data-account-id from provided html'''

    soup = BeautifulSoup(html, 'html.parser')

    form = soup.find('form')
    
    return {
        'data-account-id': form.find('button', {'data-account-id': True})['data-account-id']
    }


def get_websocket_payload(html: str) -> dict:
    ''' Searches and returns the websocket payload from provided html'''

    soup = BeautifulSoup(html, 'html.parser')
    body = soup.find('body', {'onload': True})
    next_auth_ws_init = html_lib.unescape(body['onload'])

    payload = re.search(r"'(.*)','(.*)',.*'(.*)'", next_auth_ws_init)

    return {
        'url': f"{payload.group(3)}",
        'payload': f"{payload.group(2)} {payload.group(1)}"
    }

def get_shib_idp_session_ss(html: str) -> dict:
    ''' Searches and returns shib_idp_session_ss from provided html'''

    soup = BeautifulSoup(html, 'html.parser')

    script_tag = soup.find_all('script')[-1]

    shib_idp_session_ss_match = re.search(r'", "(.*)"', str(script_tag))
    
    return shib_idp_session_ss_match.group(1)

def check_login_state(html: str) -> bool:
    ''' Searches for 'login-error' id and returns True or False with reason'''

    soup = BeautifulSoup(html, 'html.parser')

    alert = soup.find('div', {'id': 'login-error'})

    return True if alert is None else False


# __main__.py

import argparse
import sys
import json

from toledo.portal import create_session_object
from toledo.dashboard import extend_session as dashboard_extend_session
from toledo.kuloket import extend_session as kuloket_extend_session
from toledo.vivesplus import create_vivesplus_session_object
from toledo.api import create_toledo_api_object
from toledo.api import create_vivesplus_api_object


def main(args):

    option = [ arg for arg, value in args.__dict__.items() if value and arg in ('enrollments', 'upcoming', 'events', 'schedule', 'todo', 'student_info', 'dashboard', 'notices', 'get_authorization_token', 'set_authorization_token')][0]

    if args.service == 'toledo':

        session = create_session_object(
            user=args.rnumber,
            password=args.password
            )

        if option in ('todo'):

            session = dashboard_extend_session(
            portal_session=session
            )
        
        elif option in ('schedule'):

            session = kuloket_extend_session(
            portal_session=session
            )

        toledo_api = create_toledo_api_object(
            session=session
        )

        match option:

            case 'enrollments':

                output = toledo_api.get_enrollments()
            
            case 'upcoming':

                output = toledo_api.get_upcoming()

            case 'events':

                output = toledo_api.get_events(
                    type=args.events
                )

            case 'schedule':

                output = toledo_api.get_schedule()
            
            case 'todo':

                output = toledo_api.get_to_do(
                type=args.todo
                )

    elif args.service == 'vivesplus':

        if None in (args.rnumber, args.password) and args.set_authorization_token is None:

            sys.exit('vivesplus: error: the following arguments are required: --rnumber/-rn, --password/-pw')

        authorization = args.set_authorization_token if args.set_authorization_token is not None else None

        session = create_vivesplus_session_object(
            user=args.rnumber,
            password=args.password,
            authorization=authorization
        )
        
        vivesplus_api = create_vivesplus_api_object(
            session=session
        )

        match option:

            case 'schedule':

                output = vivesplus_api.get_schedule(
                    start_date=args.schedule[0],
                    end_date=args.schedule[1]
                )

            case 'student_info':

                output = vivesplus_api.get_student_info()

            case 'dashboard':

                output = vivesplus_api.get_dashboard()

            case 'notices':

                output = vivesplus_api.get_notices()

            case 'get_authorization_token':

                output = vivesplus_api.get_authorization_token()  
    
    if not args.silent:

        if option != 'get_authorization_token':

            json.dump(output, sys.stdout)
        
        else:

            with open(f'vivesplus_authorization_token_{args.rnumber[0]}.json', 'w') as f:
                json.dump(output, f, ensure_ascii=False, indent=4)

if __name__ == '__main__':

    parser = argparse.ArgumentParser(
        epilog='', usage=argparse.SUPPRESS)
    
    subparsers = parser.add_subparsers(help='help for services', dest='service', required=True)

    toledo_parser = subparsers.add_parser('toledo', help='toledo help')

    toledo_parser.add_argument(
        '--rnumber', '-rn', help='your personal rnumber', nargs=1, required=True)
    toledo_parser.add_argument(
        '--password', '-pw', help='your password', nargs=1, required=True)

    vivesplus_parser = subparsers.add_parser('vivesplus', help='vivesplus help')

    toledo_optionsgroup = toledo_parser.add_mutually_exclusive_group(required=True)
    
    toledo_optionsgroup.add_argument(
        '--enrollments', '-en', help='retrieve all your enrollments', action='store_true')
    toledo_optionsgroup.add_argument(
        '--todo', '-td', help='retrieve your to-do list (tasks or tests)', choices=['task', 'test'])
    toledo_optionsgroup.add_argument(
        '--upcoming', '-up', help='retrieve your upcoming courses', action='store_true')
    toledo_optionsgroup.add_argument(
        '--events', '-ev', help='retrieve your recent events (messages or updates)', choices=['message', 'update'])
    toledo_optionsgroup.add_argument(
        '--schedule', '-sc', help='retrieve your schedule', action='store_true')
    
    vivesplus_optionsgroup = vivesplus_parser.add_mutually_exclusive_group(required=True)

    vivesplus_optionsgroup.add_argument(
        '--schedule', '-sc', nargs=2, metavar=('start_date', 'end_date'), help='retrieve your schedule')

    vivesplus_optionsgroup.add_argument(
        '--student-info', '-si', help='retrieve your student information', action='store_true')

    vivesplus_optionsgroup.add_argument(
        '--dashboard', '-db', help='retrieve your dashboard', action='store_true')
    
    vivesplus_optionsgroup.add_argument(
        '--notices', '-no', help='retrieve your notices', action='store_true')

    vivesplus_optionsgroup.add_argument(
        '--get-authorization-token', '-gat', help='retrieve your vives plus authorization token as a JSON file', action='store_true')

    vivesplus_parser.add_argument(
        '--set-authorization-token', '-sat', help='set your vives plus authorization token by providing a JSON file')

    vivesplus_parser.add_argument(
        '--rnumber', '-rn', help='your personal rnumber', nargs=1, required=False)
    vivesplus_parser.add_argument(
        '--password', '-pw', help='your password', nargs=1, required=False)


    toledo_parser.add_argument(
        '--silent', '-s', help='surpress output', action='store_true')
    
    vivesplus_parser.add_argument(
        '--silent', '-s', help='surpress output', action='store_true')
    
    main(parser.parse_args())



import pathlib
import setuptools

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setuptools.setup(
    name="pyToledo",
    version="3.1.0",
    description="pyToledo is a Python library to interact with the common virtual learning environment for the Association KU Leuven a.k.a Toledo.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/DaanVervacke/pyToledo",
    author="Vervacke Daan",
    author_email="daan.vervacke@student.vives.be",
    license="GPLv3",
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
    ],
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=["requests", "beautifulsoup4",
                      "pkce", "soupsieve", "pyyaml", "websockets"]

)
